from colorama import Fore as f
from os.path import join
from os import listdir
from typing import List
import subprocess, re

from prompt_toolkit import PromptSession, HTML
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import WordCompleter, PathCompleter, Completer, merge_completers

class _InteractiveShell:
    def __init__(self, bin_path: str):
        self.bin_path = bin_path
        self.session = PromptSession(history=InMemoryHistory())
        self.command_history = []
        self.path_completer = PathCompleter()
        self.command_completer = WordCompleter(self.command_history, ignore_case=True)
        self.ansi_escape = re.compile(r'\x1b\[[0-9;]*m')
    
    def _execute(self, command: List[str]):
        if command[0] in listdir(self.bin_path):
            command[0] = join(self.bin_path, command[0])

        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError:
            pass
        except (KeyboardInterrupt, EOFError):
            pass
        except Exception:
            pass
    
    def _strip_ansi(self, text):
        return self.ansi_escape.sub('', text)

    def _add_to_history(self, command: str):
        stripped_command = self._strip_ansi(command.strip())
        if stripped_command and (not self.command_history or self.command_history[-1] != stripped_command):
            self.command_history.append(stripped_command)
    
    def _get_completer(self):
        return merge_completers([WordCompleter(self.command_history), self.path_completer])
    
    def start_shell(self, prompt: HTML, stop: List[str]):
        print(f"\n{f.GREEN}Emulator Enabled.{f.RESET} use \'/stop\', \'stop\', or \'deactivate\' to exit.")
        print(f"{f.YELLOW}NOTE{f.RESET}: This emulator is meant for pip, python, cleverish and any bins installed in the environment only.")
        
        # add stop commands in history.
        # and a few custom
        custom = [
            'clever init','clever shell', 'clever add', 
            'clever install', 'clever help', 'pip install', 
            'pip uninstall', 'python -m pip install --upgrade pip',
            'python -m', 'clear'
        ]
        for stop_command in stop + custom:
            self._add_to_history(stop_command)

        while True:
            try:
                self.session.completer = self._get_completer()

                user_input = self.session.prompt(prompt, auto_suggest=AutoSuggestFromHistory())

                if user_input.strip().lower() in stop:
                    print(f"{f.GREEN}exit.{f.RESET}")
                    break

                self._add_to_history(user_input)
                self._execute(user_input.split())
            except (KeyboardInterrupt, EOFError):
                print(f"{f.GREEN}exit.{f.RESET}")
                break

class TerminalEmulator:
    def __init__(self, env: str, bin_path: str) -> None:
        self.env = env
        self.bin_path = bin_path
    
    @property
    def start(self) -> None:
        stop_keywords = ['/stop', 'stop', 'deactivate', 'exit', 'quit', ':q']

        shell = _InteractiveShell(self.bin_path)
        shell.start_shell(HTML(f'(<ansiblue>{self.env})</ansiblue> <ansigreen>> </ansigreen>'), stop_keywords)
        
        


class NonEmptyDirectory(Exception):
    """`Directory Not Empty Exception`"""
    NOT_EMPTY: str

class CleverConfigurationNotFound(Exception):
    """`Clever dot toml missing Exception`"""
    INIT_FIRST: str


from os import system, chdir, getcwd, makedirs, popen, name
from os.path import exists, join
from shutil import rmtree
from pathlib import Path
import tempfile

def test_init():
    # create a temp dir
    with tempfile.TemporaryDirectory() as tempdir:
        # change to temp dir
        current = getcwd()
        makedirs(join(tempdir, 'test'))
        chdir(join(tempdir, 'test'))

        # use simple init
        system("clever init -ws")

        assert exists(join(getcwd(), '.test'))
        assert exists(join(getcwd(), '.git'))
        assert exists(join(getcwd(), 'src'))
        assert exists(join(getcwd(), 'pyproject.toml'))
        assert exists(join(getcwd(), 'clever.toml'))

        # delete the dir
        chdir(tempdir)
        rmtree(join(tempdir, 'test'))

        # create again and move
        makedirs(join(tempdir, 'test'))
        chdir(join(tempdir, 'test'))

        # test init name
        system('clever init name testing -ws')

        assert exists(join(getcwd(), '.testing'))
        assert exists(join(getcwd(), '.git'))
        assert exists(join(getcwd(), 'src'))
        assert exists(join(getcwd(), 'src', 'testing'))
        assert exists(join(getcwd(), 'pyproject.toml'))
        assert exists(join(getcwd(), 'clever.toml'))

        # delete the dir
        chdir(tempdir)
        rmtree(join(tempdir, 'test'))

        # create again and move
        makedirs(join(tempdir, 'test'))
        chdir(join(tempdir, 'test'))

        system('clever init package -ws')

        # test init package
        assert exists(join(getcwd(), '.test'))
        assert exists(join(getcwd(), '.git'))
        assert exists(join(getcwd(), 'src'))
        assert exists(join(getcwd(), 'src', 'test'))
        assert exists(join(getcwd(), 'pyproject.toml'))
        assert exists(join(getcwd(), 'clever.toml'))

        ### NOW CHECK FOR INSTALLATIONS OF build, twine\
        pip_path = Path(f'./.test') / ('bin' if name != 'nt' else 'Scripts') / 'pip'
        data = popen(f'{pip_path} list --format=freeze').read()

        if 'twine' in data and 'build' in data:
            assert True
        else:
            assert False

        # change to current
        chdir(current)
        # exits here


from os import (
    listdir,
    getcwd,
    makedirs,
    name as osname,
    environ,
    geteuid,
    unlink,
    system as run,
    popen as get_output_of
)
from os.path import join, basename, exists, dirname
from typing import Union, Tuple, List, Dict
from colorama import Fore as _
from pathlib import Path
from packaging import version
from packaging.specifiers import SpecifierSet, InvalidSpecifier
import subprocess, sys, venv, ctypes, requests

from .exceptions import NonEmptyDirectory
from .utils import Tomler, ConfigurationInterpreter, TerminalEmulator
from .utils.dotlock import DotLock

class _ConstantTextContents:
    HELP: str
    PYPROJECT_TOML: str
    DEMO_CLEVERISH_TOML: str

class _DriverMethods:
    @staticmethod
    def create_init_file_at(location: str) -> None:
        """create a `__init__.py` file at a specified location."""
        ...
    
    @staticmethod
    def create_pyproject_toml_at(location: str, name: str) -> None:
        """create initial `pyproject.toml` file at the given location with the given package name."""
        ...
    
    @staticmethod
    def create_cleverish_toml_at(location: str) -> None:
        """create a `cleverish.toml` at the specified location."""
        ...

class DriverConfiguration:
    """`Class to store all driver configurations`"""
    def __init__(self) -> None:
        """`Create a driver configuration object`"""
        ...
    
    def set_force(self, value: bool) -> None:
        """sets force to a given value"""
        ...
    
    def set_package_name(self, value: str) -> None:
        """sets package name to a given value"""
        ...
    
    def set_package_status(self, value: bool) -> None:
        """sets package status to given value."""
        ...
    
    def set_tomler(self, value: Tomler) -> None:
        """sets tomler"""
        ...
    
    def set_to_add(self, value: Union[str, List[str], None]) -> None:
        """sets new package to add."""
        ...
    
    def set_upload_directory(self, value: Union[str, None]) -> None:
        """sets the upload directory to be uploaded."""
        ...

    def set_without_shell(self, value: bool) -> None: ...
    
    @property
    def forceful_initialization(self) -> bool:
        """Force Init?"""
        ...
    
    @forceful_initialization.setter
    def forceful_initialization(self, force: bool) -> None: ...
    @forceful_initialization.deleter
    def forceful_initialization(self) -> None: ...

    @property
    def package_name(self) -> Union[str, None]:
        """`Package Name if any`"""
        ...
    
    @package_name.setter
    def package_name(self, name: Union[str, None]) -> None: ...
    @package_name.deleter
    def package_name(self) -> None: ...

    @property
    def package(self) -> bool:
        """`is it a package?`"""
        ...
    
    @package.setter
    def package(self, pkg: bool) -> None: ...
    @package.deleter
    def package(self) -> None: ...

    @property
    def tomler(self) -> Tomler:
        """`tomler if any`"""
        ...
    
    @tomler.setter
    def tomler(self, tomler: Tomler) -> None: ...
    @tomler.deleter
    def tomler(self) -> None: ...

    @property
    def to_add(self) -> Union[str, List[str], None]:
        """`to add new dependency if any.`"""
        ...
    
    @to_add.setter
    def to_add(self, value: Union[str, List[str], None]) -> None: ...
    @to_add.deleter
    def to_add(self) -> None: ...

    @property
    def upload_directory(self) -> Union[str, None]:
        """`the upload directory to look for uploading`"""
        ...

    @upload_directory.setter
    def upload_directory(self, value: Union[str, None]) -> None: ...
    @upload_directory.deleter
    def upload_directory(self) -> None: ...

    @property
    def without_shell(self) -> bool: ...
    @without_shell.setter
    def without_shell(self, value: bool): ...
    @without_shell.deleter
    def without_shell(self) -> None: ...

class Drivers:
    """`Contains Drivers for all operations in cleverish domain.`"""
    @staticmethod
    def has_admin_priviledge() -> bool:
        """`Checks if the current session is root.`"""
        ...
    
    @staticmethod
    def relaunch_with_admin_priviledge() -> None:
        """`Relaunch session with admin priviledge`"""
        ...
    
    @staticmethod
    def clever_toml() -> bool:
        """`checks if clever.toml is present`"""
        ...
    
    @staticmethod
    def help() -> None:
        """`Prints Help`"""
        ...
    
    @staticmethod
    def init(config: DriverConfiguration) -> None:
        """`Initialize a new project.`"""
        ...
    
    @staticmethod
    def create_environment(config: DriverConfiguration) -> None:
        """`create environment`"""
        ...
    
    @staticmethod
    def shell(config: DriverConfiguration) -> None:
        """`run shell.`"""
        ...
    
    @staticmethod
    def get_all_versions(name: str) -> List[str]:
        """`get all versions of a package from pypi.`"""
        ...
    
    @staticmethod
    def get_latest_compatible_version(name: str, constraint: str) -> Union[str, None]:
        """from a list of globally available versions in pypi, get the compatible one if any using the constrain. else None."""
        ...

    @staticmethod
    def resolve_dependencies(dependencies: List[Dict[str, str]]) -> List[Dict[str, str]]:
        """resolve a list of dependencies based on manifest data."""
        ...
    
    @staticmethod
    def install_dependencies(config: DriverConfiguration, explicit: bool = True, extras: List[str] = [], user_given: List[str] = []) -> Tuple[int, int]:
        """`Install the dependencies.`"""
        ...
    
    @staticmethod
    def user_install_dependencies(config:DriverConfiguration) -> Tuple[int, int]:
        """for user installation"""
        ...
    
    @staticmethod
    def _strip(string: str) -> str:
        """`strip a string`"""
        ...
    
    @staticmethod
    def _find_required_dependencies(pip_path: Path, to_check: List[str]) -> List[str]:
        """`returns dependencies of the given to_check`"""
        ...

    @staticmethod
    def add(config: DriverConfiguration) -> None:
        """`Adds a dependency`"""
        ...
    
    @staticmethod
    def add_installation(name: str, version: Union[str, None], config: DriverConfiguration) -> bool:
        """`Installs a package. returns True upon success else False.`"""
        ...
    
    @staticmethod
    def build(config: DriverConfiguration) -> None:
        """`Builds the project.`"""
        ...
    
    @staticmethod
    def upload(config: DriverConfiguration) -> None:
        """`uploads to pypi`"""
        ...


from .add_analyse import Arguments
from typing import Callable, Any, Union, List
from argpi import FetchType

class PathWays:
    """`PathWay Formatter for Arguments`"""
    def __init__(self) -> None:
        """`Create a PathWay Object`"""
        ...
    
    def register(self, name: str, func: Callable, use_argument_value: bool = False, FetchType: FetchType = FetchType.SINGULAR, *args, **kwargs) -> Any:
        """`Register an Argument for execution`
        
        Returns whatever the function returns.
        """
        ...
    
    def fetch(self, name: str, fetchtype: FetchType) -> Union[str, List[str]]:
        """`Fetch value of an argument`"""
        ...


from .tomler_ import Tomler
from ..exceptions.configuration_exceptions import ConfigurationError
from typing import Dict, List
from os.path import join, dirname as directory_name_of, basename, exists

class _ConfigurationObjectives:
    """`Internal class for all configuration paths`"""

    def __init__(self, configuration_object: Tomler) -> None:
        """`Do Not Use This Class for personal use.`"""
        ...

    @property
    def get(self) -> None:
        """`Get the configuration from Tomler Object.`
        
        This property is automatically invoked upon __init__.
        If you are looking for a reload functionality, use `reload`
        property
        """
        ...
    
    @property
    def reload(self) -> None:
        """`Reload the TOML file and it's configuration.`"""
        ...
    
    @property
    def environment_name(self) -> str:
        """`The Env Name to use for this current project.`
        
        If not provided, by default takes the `cleverish.toml` enclosing
        directory name as env name.
        """
        ...
    
    @property
    def dependency_directory(self) -> str:
        """`The Directory where the dependency for this env will be stored.`
        
        If not given, `current_dir/deps` will be used as default
        """
        ...
    
    @property
    def get_requirements_automatically(self) -> bool:
        """`Whether to Take Arguments Automatically.`"""
        ...
    
    @property
    def requirements_list(self) -> List[Dict[str, str]]:
        """`The Requirements List and Versions`"""
        ...

class ConfigurationInterpreter:
    """`Configuration Interpreter`"""
    @staticmethod
    def interpret(tomler: Tomler) -> _ConfigurationObjectives:
        """`Interpret the Configurations using a Tomler Object`"""
        ...


from .arguments import PathWays
from .utils import Tomler
from .drivers import DriverConfiguration, Drivers

from colorama import init as colorama, Fore as _
from argpi import FetchType
import sys

def clever():

    # color initialization
    colorama()
    
    function_pathways = PathWays()

    # register for each argument
    # create a configuration maintainer
    driver_configuration = DriverConfiguration()

    ###### HELP ########
    function_pathways.register('help', Drivers.help)

    ###### INIT Functionality goes here ##########

    function_pathways.register('withoutshell', driver_configuration.set_without_shell, False, FetchType.SINGULAR, True)

    ## get if force
    function_pathways.register('force', driver_configuration.set_force, False, FetchType.SINGULAR, True)

    ## get if package Name provided, if provided automatically set package to True too
    function_pathways.register('name', driver_configuration.set_package_name, True, FetchType.SINGULAR)
    
    ## get if package_name not given but it is a package.
    function_pathways.register('package', driver_configuration.set_package_status, False, FetchType.SINGULAR, True)

    ## do init
    function_pathways.register('init', Drivers.init, False, FetchType.SINGULAR, driver_configuration)
    # If this skips, go to the next functionality.

    ### This Following check is mandatory for all other commands. ####
    # Check this if anything except init.
    if not Drivers.clever_toml():
        print(f"{_.RED}Cleverish{_.RESET}: cleverish.toml file not present in current directory.")
        sys.exit(1)
    
    driver_configuration.set_tomler(Tomler())
    
    ######## Shell Functionality goes here. ######

    function_pathways.register('shell', Drivers.shell, False, FetchType.SINGULAR, driver_configuration)

    ######## Install dependencies ###############

    function_pathways.register('install', Drivers.user_install_dependencies, False, FetchType.SINGULAR, driver_configuration)

    ######## Add Functionality goes here ##########

    # have to get the package_name to add.
    function_pathways.register('add', driver_configuration.set_to_add, True, FetchType.TILL_LAST)
    
    # add them now
    function_pathways.register('add', Drivers.add, False, FetchType.SINGULAR, driver_configuration)

    ########### BUILD Functionality ############

    function_pathways.register('build', Drivers.build, False, FetchType.SINGULAR, driver_configuration)

    ########### UPLOAD Functionality ##############

    function_pathways.register('upload', driver_configuration.set_upload_directory, True, FetchType.SINGULAR)

    function_pathways.register('upload', Drivers.upload, False, FetchType.SINGULAR, driver_configuration)



from toml import load as load_configuration, dump as dump_configuration
from typing import Any as AnyEntryData, Union as Either, List
from os.path import exists, isdir, isfile, join, expanduser, expandvars
from os import getcwd as current_working_directory

from ..exceptions.tomler_exceptions import TomlDirectoryError, TomlFileError, TomlerError

class Tomler:
    def __init__(self, configuration_directory: str = current_working_directory(), configuration_name: List[str] = ['clever', 'Clever']) -> None:
        self.configuration_directory = expanduser(expandvars(configuration_directory))

        if self.configuration_directory.startswith('./'):
            self.configuration_directory = join(current_working_directory(), self.configuration_directory[2:])
        
        if not isdir(self.configuration_directory):
            raise TomlDirectoryError(TomlDirectoryError.NOT_A_DIR.format(self.configuration_directory))
        
        Configuration_Check = False
        self.configuration_path = None

        for __configuration__ in configuration_name:
            self.configuration_path = join(self.configuration_directory, f'{__configuration__}.toml')

            if exists(self.configuration_path):
                Configuration_Check = True
            
            if not isfile(self.configuration_path):
                Configuration_Check = False
                self.configuration_path = None
            
            if Configuration_Check:
                break
        
        if self.configuration_path is None:
            raise TomlFileError(f"Couldn't Find Matching Configuration files: {configuration_name}")
        
        self.load_configuration
    
    @property
    def filepath(self) -> str:
        return self.configuration_path
    
    @property
    def load_configuration(self) -> None:
        with open(self.configuration_path, 'r+') as reference:
            self.configuration_data = load_configuration(reference)
    
    @property
    def dump_configuration(self) -> None:
        with open(self.configuration_path, 'w+') as reference:
            dump_configuration(self.configuration_data, reference)
    
    def add_or_modify_entry(self, root: str, value: AnyEntryData) -> None:
        keys = root.split('.')
        data = self.configuration_data

        # Traverse the keys except the last one, ensuring nested dictionaries are created
        for key in keys[:-1]:
            data = data.setdefault(key, {})
        
        data[keys[-1]] = value

        self.dump_configuration
    
    def remove_entry(self, key: str, does_not_exists_ok: bool = False) -> None:
        keys = key.split('.')
        data = self.configuration_data
        try:
            for k in keys[:-1]:
                data = data[k]
            
            del data[keys[-1]]
            self.dump_configuration
        except KeyError:
            if not does_not_exists_ok:
                raise TomlerError(f"Key removal failed for \'{key}\', Does not exist.")
    
    def fetch_entry(self, key: str) -> Either[AnyEntryData, None]:
        keys = key.split('.')
        data = self.configuration_data
        try:
            for k in keys:
                data = data[k]
            return data
        except KeyError:
            return None


class ArgumentDefinition:
    __arguments__ = {
        # Add Command
        "add": "-a",

        # shell command
        "shell": '-sh',

        # Init command
        "init": "-i",
        "force": "-f",
        "name": "-n",
        "package": "-pkg",
        "withoutshell": "-ws",

        # Install commad
        "install": "-in",

        # Build Command,
        "build": "-b",

        # upload command,
        "upload": "-u",

        # Help Command
        "help": "-h",
    }


from .tomler_exceptions import (
    TomlerError,
)

from .configuration_exceptions import ConfigurationError
from .drivers import NonEmptyDirectory, CleverConfigurationNotFound

__all__ = [
    # Tomler
    "TomlerError",

    # Configuration Exception
    "ConfigurationError",

    # Driver Exceptions
    "NonEmptyDirectory",
    "CleverConfigurationNotFound"
]


from .add_analyse import Arguments
from typing import Callable, Any, Union, List
from argpi import FetchType

class PathWays:
    def __init__(self) -> None:
        self.arguments = Arguments().analyse_and_get
    
    def register(self, name: str, func: Callable, use_argument_value: bool = False, FetchType: FetchType = FetchType.SINGULAR, *args, **kwargs) -> Any:
        if self.arguments.__there__(name):
            if use_argument_value:
                return func(self.fetch(name, FetchType))
            else:
                return func(*args, **kwargs)
    
    def fetch(self, name: str, fetchtype: FetchType) -> Union[str, List[str]]:
        return self.arguments.__fetch__(name, fetchtype)


from .definitions import ArgumentDefinition
from argpi import Arguments as argpi, ArgumentDescription

class Arguments:
    """`Argument Parser`"""
    def __init__(self) -> None: ...
    @property
    def analyse_and_get(self) -> argpi: ...


from toml import load as load_configuration, dump as dump_configuration
from typing import Any as AnyEntryData, Union as Either, List
from os.path import exists, isdir, isfile, join, expanduser, expandvars
from os import getcwd as current_working_directory

from ..exceptions.tomler_exceptions import TomlDirectoryError, TomlFileError, TomlerError

class Tomler:
    """`Tomler Class helps with Cleverish.Toml file for loading or saving configurations`"""
    
    def __init__(
            self,
            configuration_directory: str = current_working_directory(),
            configuration_name: List[str] = ['clever', 'Clever'],
        ) -> None:
        """`Tomler Object Creation`

        Raises TomlDirectoryError or TomlFileError based on check conditions.
        
        #### Parameter Description

        - `configuration_directory`: The Directory to look for configuration toml.
        - `configuration_name`: The eligible names for the configuration file.
        """
        ...
    
    @property
    def filepath(self) -> str:
        """`Returns the Toml FilePath`"""
        ...
    
    @property
    def load_configuration(self) -> None:
        """`Load Configuration From the Given Paramters in __init__.
        This property is automatically invoked in __init__.
        Use it to reload the Data.`"""
        ...
    
    @property
    def dump_configuration(self) -> None:
        """`Save The Configuration Data To The Configuration File.
        This property is automatically invoked in the internal functions while adding,
        changing or deleting entries. Use it to re-save it.`"""
        ...
    
    def add_or_modify_entry(self, root: str, value: AnyEntryData) -> None:
        """`Add or modify any entry.`
        
        Creates an entry if Not present.

        #### Parameter Description

        - `root`: The Key value to which data will be stored.

        For Example:

        Let us take the below written TOML Configuration as an example

        > [config]  
        > auto-detect = true

        To change this `auto-detect` value, use the following python code:

        ```
        >>> from cleverish.utils import Tomler
        >>> tomler = Tomler(...)
        >>> tomler.add_or_modify_entry(
        ...     root='config.auto-detect',
        ...     value=True
        ... )
        ```
        """
        ...
    
    def remove_entry(self, key: str, does_not_exists_ok: bool = False) -> None:
        """`Remove Any Entry`
        
        Raises `TomlerError` if `does_not_exists_ok` parameter is set to False, Default is False.

        For the following toml config:

        > [config]
        > auto-detect = true

        to remove `auto-detect`, use the following tomler code in python

        ```
        >>> from cleverish.utils import Tomler
        >>> tomler = Tomler(...)
        >>> tomler.remove_entry(
        ...     key='config.auto-detect'
        ... )
        ```
        """
        ...
    
    def fetch_entry(self, key: str) -> Either[AnyEntryData, None]:
        """`Fetch Any Key Value.`
        
        Returns `None` if key does not exist.
        """
        ...



from typing import List, Dict
from json import dump as dump_as_json, load as load_from_json

class DotLock:
    @staticmethod
    def create(dependencies: List[Dict[str, str]], name: str = 'clever') -> None:
        with open(name + '.lock', 'w+') as reference:
            dump_as_json(dependencies, reference, indent=4)
        
    @staticmethod
    def load(name: str = 'clever') -> List[Dict[str, str]]:
        with open(name + '.lock', 'r+') as reference:
            return load_from_json(reference)
    


from os import (
    listdir,
    getcwd,
    makedirs,
    name as osname,
    environ,
    unlink,
    system as run,
    popen as get_output_of
)

if osname != 'nt':
    from os import geteuid

from os.path import join, basename, exists, dirname
from typing import Union, Tuple, List, Dict
from pathlib import Path
from colorama import Fore as _
from packaging import version
from packaging.specifiers import SpecifierSet, InvalidSpecifier
import subprocess, sys, venv, ctypes, requests

from .exceptions import NonEmptyDirectory
from .utils import Tomler, ConfigurationInterpreter, TerminalEmulator
from .utils.dotlock import DotLock

class _ConstantTextContents:

    HELP = f"""{_.GREEN},-----------,{_.RESET} ,------------------,
{_.GREEN}|{_.RESET} Help Text {_.GREEN}|{_.RESET}-| {_.BLUE}Cleverish{_.RESET} v{_.RED}0.0.1{_.RESET} |
{_.GREEN}'-----------'{_.RESET} '------------------'

Credits: Soumyo Deep Gupta (d33p0st (GitHub)) 2024
License: MIT

Syntax: {_.BLUE}clever{_.RESET} [OPTIONS] [OPTION-VALUES] [SUB OPTIONS]

{_.YELLOW}[OPTIONS]{_.RESET}
help (-h)            :  Show this help text and exit.

init (-i)            :  Initialize an empty project.
      {_.YELLOW}[SUB OPTIONS]{_.RESET} (When using init, these can be used with it)
      force (-f)     :  If the current directory is non-empty, init
                        will raise NonEmptyDirectory Error. To force
                        it, use this flag.
    
      package (-pkg) :  Use this flag if you want to create a package
                        type project (src/<package_name>/__init__.py)
                        else it will be (src/__init__.py) type structure.
    
      name (-n)      :  Set the package name. If not provided, enclosing
                        directory name will be used as package name.
                        NOTE: If (-pkg) flag is not given but (-n) is given
                        (-pkg) flag will be auto-added.

add (-a)             :  Add a dependency.
      {_.YELLOW}[OPTION-VALUES]{_.RESET}
      This OPTION can take values in some formats I will add later.
"""

    PYPROJECT_TOML = """# Generated using Cleverish

[build-system]
requires = ["setuptools>=42", "wheel"] # Auto generated with cleverish, you can change it as per your needs
build-backend = "setuptools.build_meta" # Auto generated with cleverish, you can change it as per your needs
    
[project]
name = "{}"
version = "0.0.1"
description = "Auto Generated Text using Cleverish"
readme = { file = "README.md", content-type = "text/markdown"}
license = { file = "LICENSE" }
"""

    DEMO_CLEVERISH_TOML = """
[config]
environment_name = \'{}\'

[requirements]
dependency_directory = 'deps'
get-automatically = false
list = []
"""

class DriverConfiguration:
    def __init__(self) -> None:
        self._forceful_initialization: bool = False
        self._package_name: Union[str, None] = None
        self._package: bool = False
        self._ws: bool = False
        self._tomler: Tomler
        self._to_add: Union[str, List[str], None,] = None
        self._upload_directory: Union[str, None] = None
    
    def set_force(self, value: bool) -> None:
        self.forceful_initialization = value
    
    def set_package_name(self, value: Union[str, None]) -> None:
        self.package_name = value
    
    def set_package_status(self, value: bool) -> None:
        self.package = value
    
    def set_tomler(self, value: Tomler) -> None:
        self._tomler = value
    
    def set_to_add(self, value: Union[str, List[str], None]) -> None:
        self._to_add = value
    
    def set_upload_directory(self, value: Union[str, None]) -> None:
        self._upload_directory = value
    
    def set_without_shell(self, value: bool) -> None:
        self._ws = value
    
    @property
    def forceful_initialization(self) -> bool:
        return self._forceful_initialization
    
    @forceful_initialization.setter
    def forceful_initialization(self, force: bool) -> None:
        self._forceful_initialization = force
    
    @forceful_initialization.deleter
    def forceful_initialization(self) -> None:
        self.forceful_initialization = False
    
    @property
    def package_name(self) -> Union[str, None]:
        return self._package_name
    
    @package_name.setter
    def package_name(self, name: Union[str, None]) -> None:
        self._package_name = name
        if self._package_name:
            self.package = True
    
    @package_name.deleter
    def package_name(self) -> None:
        self.package_name = None
    
    @property
    def package(self) -> bool:
        return self._package
    
    @package.setter
    def package(self, pkg: bool) -> None:
        self._package = pkg
    
    @package.deleter
    def package(self) -> None:
        self.package = False
    
    @property
    def tomler(self) -> Tomler:
        return self._tomler
    
    @tomler.setter
    def tomler(self, tomler: Tomler) -> None:
        self._tomler = tomler
    
    @tomler.deleter
    def tomler(self) -> None:
        self._tomler = None
    
    @property
    def to_add(self) -> Union[str, List[str], None]:
        return self._to_add
    
    @to_add.setter
    def to_add(self, value: Union[str, List[str], None]) -> None:
        self._to_add = value
    
    @to_add.deleter
    def to_add(self) -> None:
        self._to_add = None
    
    @property
    def upload_directory(self) -> Union[str, None]:
        return self._upload_directory
    
    @upload_directory.setter
    def upload_directory(self, value: Union[str, None]) -> None:
        self._upload_directory = value
    
    @upload_directory.deleter
    def upload_directory(self) -> None:
        self._upload_directory = None
    
    @property
    def without_shell(self) -> bool:
        return self._ws
    
    @without_shell.setter
    def without_shell(self, v: bool) -> None:
        self._ws = v
    
    @without_shell.deleter
    def without_shell(self) -> None:
        self._ws = False

class _DriverMethods:
    @staticmethod
    def create_init_file_at(location: str) -> None:
        with open(join(location, '__init__.py'), 'w+') as reference:
            reference.write("# Init File - Auto Generated Through Cleverish.")
    
    @staticmethod
    def create_pyproject_toml_at(location: str, name: str) -> None:
        with open(join(location, 'pyproject.toml'), 'w+') as reference:
            reference.write(_ConstantTextContents.PYPROJECT_TOML.replace("{}", name))
    
    @staticmethod
    def create_cleverish_toml_at(location: str, name: str) -> None:
        with open(join(location, 'clever.toml'), 'w+') as reference:
            reference.write(_ConstantTextContents.DEMO_CLEVERISH_TOML.replace("{}", name))

class Drivers:
    @staticmethod
    def has_admin_priviledge() -> bool:
        if osname != 'nt':
            return geteuid() == 0
        else:
            try:
                return ctypes.windll.shell32.IsUserAnAdmin()
            except AttributeError:
                return False
    
    @staticmethod
    def relaunch_with_admin_priviledge() -> None:
        if osname != 'nt':
            try:
                subprocess.check_call(['sudo'] + sys.argv)
            except subprocess.CalledProcessError:
                print(f"{_.RED}Failed{_.RESET} to run with sudo.\nRun the following to run it as root.\n{_.YELLOW}sudo{_.RESET} " + ' '.join(sys.argv))
                sys.exit(1)
        else:
            try:
                ctypes.windll.shell32.shellExecute(None, "runas", sys.executable, " ".join(sys.argv), None, 1)
            except Exception:
                print(f"{_.RED}Failed{_.RESET} to elevate priviledge automatically. Run it in administrator cmd/powershell.")
                sys.exit(1)
        
        sys.exit(0)
    
    @staticmethod
    def clever_toml() -> bool:
        files = listdir(getcwd())
        return 'clever.toml' in files or 'Clever.toml' in files

    @staticmethod
    def help() -> None:
        print(_ConstantTextContents.HELP)
        sys.exit(0)
    
    @staticmethod
    def init(config: DriverConfiguration) -> None:
        # # Initiate root if not root.
        # if not Drivers.has_admin_priviledge():
        #     Drivers.relaunch_with_admin_priviledge()

        # Get Contents of the current directory if not force
        if not config.forceful_initialization:
            files_in_current_directory = listdir(getcwd())
            
            if '.git' in files_in_current_directory:
                files_in_current_directory.remove('.git')
            
            if files_in_current_directory:
                raise NonEmptyDirectory(NonEmptyDirectory.NOT_EMPTY.format(getcwd()) + " Init Failed.")
        
        # create src
        makedirs(join(getcwd(), 'src'), exist_ok=True)

        package_name = None

        # create a directory under src if package else just create __init__.py
        if not config.package:
            # Not a package
            _DriverMethods.create_init_file_at(join(getcwd(), 'src'))
        else:
            # IT is a package.
            # check if name is provided
            package_name: str
            if config.package_name:
                package_name = config.package_name
                # Create the package name dir
                makedirs(join(getcwd(), 'src', config.package_name), exist_ok=True)
                # create init file inside
                _DriverMethods.create_init_file_at(join(getcwd(), 'src', config.package_name))
            else:
                # If name not provided,
                # get name of the current dir.
                package_name = basename(getcwd())
                # create the package name dir
                makedirs(join(getcwd(), 'src', package_name), exist_ok=True)
                _DriverMethods.create_init_file_at(join(getcwd(), 'src', package_name))
        
        # src -> ?Package_name_dir? -> __init__.py COMPLETE HERE

        # create a pyproject.toml
        _DriverMethods.create_pyproject_toml_at(getcwd(), package_name if package_name else basename(getcwd()))

        # create demo cleverish.toml
        _DriverMethods.create_cleverish_toml_at(getcwd(), package_name if package_name else basename(getcwd()))

        # If not a git repository, run git init
        if not exists(join(getcwd(), '.git')):
            try:
                subprocess.check_call(['git', 'init'], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
            except subprocess.CalledProcessError:
                print(f"{_.RED}Git repository at this location cannot be created.{_.RESET} Create manually if needed.")
        
        print(f"{_.GREEN}Succesfully created{_.RESET} empty project at current directory.")

        # tomler needs to be created and added to config for env
        new_config = DriverConfiguration()
        new_config.set_force(config.forceful_initialization)
        new_config.set_package_name(config.package_name)
        new_config.set_package_status(config.package)
        new_config.tomler = Tomler()
        new_config.set_without_shell(config.without_shell)
        Drivers.create_environment(config=new_config) # create env too
        sys.exit(0)
    
    @staticmethod
    def create_environment(config: DriverConfiguration) -> None:
        print(f"{_.YELLOW}Creating Environment...{_.RESET}", end='\r')
        # get tomler
        tomler = config.tomler
        interpreted_tomler = ConfigurationInterpreter.interpret(tomler)

        # create environment
        env_path = Path(f'./.{interpreted_tomler.environment_name}')
        if not env_path.exists() or config.forceful_initialization:
            venv.create(env_path, with_pip=True)
        else:
            print("                                                  ", end='\r')
            print(f"{_.YELLOW}Environment Exists{_.RESET}")
            print(f"Run: {_.BLUE}clever{_.RESET} shell")
            sys.exit(0)
        
        if osname == 'nt':
            activator = env_path / 'Scripts' / 'activate'
        else:
            activator = env_path / 'bin' / 'activate'
        
        print(f"{_.GREEN}Environment created as per configuration.{_.RESET}")

        print(f"{_.YELLOW}Installing requirements (if any) ...{_.RESET}", end='\r')

        skipped, total = Drivers.install_dependencies(config, False, ['build', 'twine'], []) # later on add cleverish
        print(f"                                                                  ", end='\r')
        
        if skipped == total:
            print(f"{_.RED}Failed{_.RESET} to install dependencies.")
        elif skipped > 0:
            print(f"requirements {_.GREEN}Installed{_.RESET} {_.YELLOW}(partially){_.RESET}")
            print(f"{_.YELLOW}Skipped{_.RESET} {skipped} dependencies.")
        else:
            print(f"requirements {_.GREEN}Installed{_.RESET}")

        if not config.without_shell:
            TerminalEmulator(env=interpreted_tomler.environment_name, bin_path=dirname(str(activator))).start
    
    @staticmethod
    def shell(config: DriverConfiguration) -> None:
        itomler = ConfigurationInterpreter.interpret(config.tomler)
        env_path = Path(f'./.{itomler.environment_name}')
        
        if not env_path.exists():
            print(f"({_.BLUE}{itomler.environment_name}{_.RESET}) {_.RED}Environment Does Not Exist{_.RESET}")
            print(f"Run: {_.BLUE}clever{_.RESET} init")
            print(f"For more information: {_.BLUE}clever{_.RESET} help")
            sys.exit(1)
        
        if osname == 'nt':
            bin_p = env_path / 'Scripts'
        else:
            bin_p = env_path / 'bin'
        
        TerminalEmulator(env=itomler.environment_name, bin_path=bin_p).start
    
    @staticmethod
    def get_all_versions(name: str) -> List[str]:
        url = f"https://pypi.org/pypi/{name}/json"
        try:
            response = requests.get(url)
            try:
                response.raise_for_status()
            except requests.HTTPError:
                print(f"{_.RED}Failed{_.RESET} to resolve dependency for {name}.")
                sys.exit(1)
            data = response.json()
            return sorted(data["releases"].keys(), key=version.parse, reverse=True)
        except requests.exceptions.RequestException:
            print(f"{_.RED}Failed{_.RESET} to resolve dependency for {name}.")
            sys.exit(1)
    
    @staticmethod
    def get_latest_compatible_version(name: str, constraint: str) -> Union[str, None]:
        available = Drivers.get_all_versions(name)
        try:
            compatible = [v for v in available if version.parse(v) in SpecifierSet(constraint)]
        except InvalidSpecifier:
            print(f"{_.RED}Invalid Specifier{_.RESET}: \'{constraint}\' for {name}.")
            sys.exit(1)
        
        return compatible[0] if compatible else None

    @staticmethod
    def resolve_dependencies(dependencies: List[Dict[str, str]]) -> List[Dict[str, str]]:
        resolved = []
        for dep in dependencies:
            name = dep.get('name', None)
            constraint = dep.get('version', None)
            if not name:
                resolved.append(dep) # will be skipped later in "install_dependencies"
                continue

            if not constraint:
                resolved.append(dep) # name but no version. will be handled in "install_dependencies"
                continue

            specific_v = Drivers.get_latest_compatible_version(name, constraint)
            resolved.append({"name": name, "version": specific_v})
        return resolved

    @staticmethod
    def install_dependencies(config: DriverConfiguration, explicit: bool = True, extras: List[str] = [], user_given: List[str] = []) -> Tuple[int, int]:
        itomler = ConfigurationInterpreter.interpret(config.tomler)
        if not exists(join(getcwd(), 'clever.lock')):
            dependencies = Drivers.resolve_dependencies(itomler.requirements_list)
            for x in dependencies:
                user_given.append(x['name'])
        else:
            dependencies = DotLock.load()
        
        for x in extras:
            dependencies.append({"name": x})
        
        env_path = Path(f'./.{itomler.environment_name}') / "bin" if osname != 'nt' else Path(f'./.{itomler.environment_name}') / "Scripts"
        pip_path = env_path / 'pip'
        py_path = env_path / 'python'

        skip_count = 0
        for dependency in dependencies:
            name = dependency.get('name', None)
            version = dependency.get('version', None)

            if not name:
                skip_count += 1
                continue

            list_of_i = get_output_of(f"{pip_path} list --format=freeze").readlines()
            v_to_check = None
            for i in range(len(list_of_i)):
                if name in list_of_i[i]:
                    list_of_i[i] = list_of_i[i].replace('\n', '')
                    v_to_check = list_of_i[i].split('==')[1]
                    break
            
            # if not pip and (either (current_installed_version != version) or (no current_installed_version))
            if name != 'pip' and ((v_to_check and v_to_check != version) or (not v_to_check)):
                icommand = f"{pip_path} install {name}" + ("" if version is None else f"=={version}")

                try:
                    subprocess.run(icommand.split(), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                    if explicit:
                        print(f"{_.GREEN}Added{_.RESET} {name}.")
                except subprocess.CalledProcessError:
                    print(f"{_.RED}Failed{_.RESET} to install {name}. {_.YELLOW}Skipping{_.RESET}")
                    skip_count += 1
                    continue
            elif {"name": "pip", "version": Drivers.get_all_versions('pip')[0]} not in dependencies:
                try:
                    subprocess.run(f"{py_path} -m pip install --upgrade pip".split(), check=True, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
                    print(f"{_.GREEN}Upgraded{_.RESET} pip -> {Drivers.get_all_versions('pip')[0]}")
                except subprocess.CalledProcessError:
                    print(f"{_.RED}Failed{_.RESET} to upgrade pip. {_.GREEN}An upgrade is available{_.RESET}")
                    sys.exit(0)
            else:
                skip_count += 1
        
        # get list of installed versions from pip
        list_of_installed = get_output_of(f"{pip_path} list --format=freeze").readlines()
        dependency_list_complete = []

        for i in range(len(list_of_installed)):
            list_of_installed[i] = list_of_installed[i].replace('\n', '')
            name, version = tuple(list_of_installed[i].split('=='))
            if name not in user_given:
                continue
            dependency_list_complete.append(
                {
                    "name": name,
                    "version": version,
                    "dependencies": Drivers._find_required_dependencies(pip_path, [name])
                }
            )

        DotLock.create(dependencies=dependency_list_complete)
        
        tomler = config.tomler
        itomler = ConfigurationInterpreter.interpret(tomler)
        current_requirement_menu = tomler.fetch_entry('requirements')
        current_requirement_menu['list'] = dependency_list_complete
        tomler.add_or_modify_entry('requirements', current_requirement_menu)

        return skip_count, len(dependencies)
    
    @staticmethod
    def user_install_dependencies(config: DriverConfiguration) -> Tuple[int, int]:
        return Drivers.install_dependencies(config, True, [], [x['name'] for x in ConfigurationInterpreter.interpret(config.tomler).requirements_list])

    @staticmethod
    def _strip(string: str) -> str:
        return string.strip()
    
    @staticmethod
    def _find_required_dependencies(pip_path: Path, to_check: List[str]) -> List[str]:
        # consider all the packages provided are already installed.
        deps = []
        for each in to_check:
            data = get_output_of(f'{pip_path} show {each}').readlines()
            for line in data:
                if 'Requires' in line:
                    line = line.replace('\n', '')
                    line = line.split(':')[1].strip()
                    if line != '':
                        deps_here = list(map(Drivers._strip, line.split(',')))
                        for dep in deps_here:
                            if dep not in deps:
                                deps.append(dep)

        return deps

    @staticmethod
    def _resolve_package_name_and_divide(name: str) -> Tuple[str, Union[str, None]]:
        v = None
        if "=" not in name:
            v = None
        else:
            n = ''
            for i in range(len(name)):
                if name[i].isalpha():
                    n += name[i]
                else:
                    v = name[i:]
                    break
            name = n
        
        return name, v
    
    @staticmethod
    def _remove_signs_from_version(version: str) -> str:
        while not version[0].isnumeric():
            version = version[1:]
        return version
    
    @staticmethod
    def add(config: DriverConfiguration) -> None:
        packages: List[str] = config.to_add
        tomler = config.tomler
        itomler = ConfigurationInterpreter.interpret(tomler)

        list_from_toml = itomler.requirements_list

        to_install = []
        ignored = []
        shadow_installed = []

        # packages are the ones i need to add
        for package in packages:
            pkg_name, pkg_version = Drivers._resolve_package_name_and_divide(package)
            
            # none_check
            there_is_none = pkg_version is None

            if pkg_version is not None:
                pkg_version = Drivers._remove_signs_from_version(pkg_version)
            

            # in the list from toml, check if the current package to install,
            # is already installed or not
            breaker = False
            found_in_toml = False
            for element_from_toml in list_from_toml:
                
                # if already installed
                # check if versions are same or not
                if element_from_toml['name'] == pkg_name:
                    # if the name is found, but given pkg version is None,
                    if there_is_none:
                        ignored.append((pkg_name, pkg_version, f'{pkg_name} present.'))
                        breaker = True
                        break
                    
                    # try to get version from the toml
                    version_from_toml = element_from_toml.get('version', None)

                    # If version found in toml and check if pkg_to_install does not have the version same as the version in the toml
                    # if they dont, i need to install it
                    if version_from_toml is not None and version_from_toml != pkg_version:
                        to_install.append((pkg_name, pkg_version))
                    
                    elif version_from_toml is not None and version_from_toml == pkg_version:
                        # if version found in toml and if pkg to install has the version same as
                        # the toml version then it needs to be ignore, it is already installed
                        ignored.append((pkg_name, pkg_version, 'present'))

                    elif version_from_toml is None:
                        # If version is not there in the toml.
                        # Try to find it from the env_pip freeze list

                        pip_path = Path(f'./.{itomler.environment_name}') / ('bin' if osname != 'nt' else 'Scripts') / 'pip'

                        # get output of
                        # pip list --format=freeze
                        list_from_freeze = get_output_of(f'{pip_path} list --format=freeze').readlines()

                        # from the pip list (currently installed), check if the list contains version which
                        # i want to install.
                        found = False
                        for element_from_freeze in list_from_freeze:
                            # if the package name is in the element (package_name==version) (format)
                            if pkg_name in element_from_freeze:
                                found = True
                                a, version_from_freeze = Drivers._resolve_package_name_and_divide(element_from_freeze.replace('\n', ''))

                                # check if the version from freeze is the same as i want to install
                                if version_from_freeze == pkg_version:
                                    # if same, ignore it
                                    ignored.append((pkg_name, pkg_version, 'present'))
                                else:
                                    to_install.append((pkg_name, pkg_version))
                                
                                break
                        
                        # If not found in the freeze list, add it for install
                        if not found:
                            to_install.append((pkg_name, pkg_version))
                    
                    breaker = True
                    break
            
            if there_is_none and breaker:
                continue
                # Continue if package version is not provided
                # and breaker is triggered, meaning found
                
            # Breaker will be only true if name is found and version is found
            # and decision is made on ignorance or installation
            if breaker:
                break
            else:
                # If breaker is false then decision is not made.
                # does not exist and needs to be installed.

                # OK Here it is possible that it was not found in toml but is installed.
                # maybe it was installed by pip installed.

                pip_path = Path(f'./.{itomler.environment_name}') / ('bin' if osname != 'nt' else 'Scripts') / 'pip'

                # get output of
                # pip list --format=freeze
                list_from_freeze = get_output_of(f'{pip_path} list --format=freeze').readlines()

                # from the pip list (currently installed), check if the list contains version which
                # i want to install.
                found = False
                for element_from_freeze in list_from_freeze:
                    # if the package name is in the element (package_name==version) (format)
                    if pkg_name in element_from_freeze:

                        found = True
                        a, version_from_freeze = Drivers._resolve_package_name_and_divide(element_from_freeze.replace('\n', ''))
                        version_from_freeze = Drivers._remove_signs_from_version(version_from_freeze)

                        # If the pkg_version is None
                        # and there is one installed using pip
                        if pkg_version is None:
                            # IT means it is shadown installed. needs to be added
                            shadow_installed.append((pkg_name, version_from_freeze))
                            break

                        # check if the version from freeze is the same as i want to install
                        if version_from_freeze == pkg_version:
                            # if same, still, it is present in pip but not in toml
                            shadow_installed.append((pkg_name, version_from_freeze))
                        else:
                            # through pip it was known that it is present, but the pkg_version is given different than
                            # what is installed. Therefore, install it.
                            to_install.append((pkg_name, pkg_version))
                        
                        # since found, break it.
                        break
                
                # if not found, install it.
                if not found:
                    to_install.append((pkg_name, pkg_version))

        # For each package to install.
        for package in to_install:
            # if installation is done.
            if Drivers.add_installation(package[0], package[1], config):
                # check the list from tomler
                found = False
                # in the list from toml
                for i in range(len(list_from_toml)):
                    # if the entry exists
                    if list_from_toml[i]['name'] == package[0]:
                        # change the version
                        list_from_toml[i]['version'] = package[1]
                        found = True
                        break
                
                # If found, then version is changed/added, move on to next package
                if found:
                    continue
                else:
                    # If not found
                    # just add an entry
                    list_from_toml.append({"name": package[0], "version": package[1]})
            else:
                # If installation fails, add it in the ignored.
                ignored.append((package[0], package[1], 'failed'))

        for i in range(len(shadow_installed)):
            data = {"name": shadow_installed[i][0], "version": shadow_installed[i][1]}
            if not data in list_from_toml:
                list_from_toml.append(data)
            ignored.append(([shadow_installed[i][0], shadow_installed[i][1], 'present']))

        # Update the list from toml with propert version
        for x in list_from_toml:
            x['version'] = Drivers.get_version_from_freeze(x['name'], pip_path = Path(f'./.{itomler.environment_name}') / ('bin' if osname != 'nt' else 'Scripts') / 'pip')
            x['dependencies'] = Drivers._find_required_dependencies(pip_path, [x['name']])

        # update toml file
        requirements_dict: Dict = tomler.fetch_entry('requirements')
        requirements_dict['list'] = list_from_toml
        tomler.add_or_modify_entry('requirements', requirements_dict)

        # CREATE DOTLOCK
        DotLock.create(list_from_toml)

        # skipped status
        print(f"{_.RED if len(ignored) > 0 else _.GREEN}Skipped{_.RESET} {len(ignored)} : \n"+ '\n'.join(list(map(str, ignored))))
    
    @staticmethod
    def get_version_from_freeze(name: str, pip_path: Path) -> Union[str, None]:
        data = get_output_of(f"{pip_path} list --format=freeze").readlines()

        for line in data:
            if name in line:
                line = line.replace('\n', '')
                __, version = Drivers._resolve_package_name_and_divide(line)
                return Drivers._remove_signs_from_version(version)
        
        return None


    @staticmethod
    def add_installation(name: str, version: Union[str, None], config: DriverConfiguration) -> bool:
        tomler = config.tomler
        tomler.load_configuration

        itomler = ConfigurationInterpreter.interpret(tomler)
        pip_path = Path(f'./.{itomler.environment_name}') / ('bin' if osname != 'nt' else 'Scripts') / 'pip'

        command = f"{pip_path} install {name}" + (f"=={version}" if version else '')

        print(f"{_.YELLOW}Installing{_.RESET} {name}" + (f"=={version}" if version else "" + " ..."), end='\r')
        try:
            subprocess.run(command.split(), check=True, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
            print(f"                                                                         ", end='\r')
            print(f"{_.GREEN}Installed{_.RESET} {name}" + (f"=={version}." if version else "."))
            return True
        except subprocess.CalledProcessError:
            print("                                                          ", end='\r')
            print(f"{_.RED}Failed{_.RESET} to install {name}" + (f"=={version}" if version else ""))
            return False
    
    @staticmethod
    def build(config: DriverConfiguration) -> None:
        itomler = ConfigurationInterpreter.interpret(config.tomler)

        py_path = Path(f'./.{itomler.environment_name}') / ('bin' if osname != 'nt' else 'Scripts') / 'python'
        command = f"{py_path} -m build"
        print(f"{_.YELLOW}Building.{_.RESET}", end='\r')
        try:
            subprocess.run(command.split(), check=True, stdout=subprocess.DEVNULL)
            print(f"                            ", end='\r')
            print(f"{_.GREEN}Built.{_.RESET} ./dist")
        except subprocess.CalledProcessError:
            print("                           ", end='\r')
            print(f"Build {_.RED}Failure{_.RESET}")
            sys.exit(1)

        sys.exit(0)
    
    @staticmethod
    def upload(config: DriverConfiguration) -> None:
        upload_dir = config.upload_directory
        if upload_dir is None:
            print(f"Please provide a {_.RED}Upload Directory Path{_.RESET}.\nFor help run: {_.BLUE}clever{_.RESET} help")
            sys.exit(1)
        
        twine_path = Path(f'./.{ConfigurationInterpreter.interpret(config.tomler).environment_name}') / ('bin' if osname != 'nt' else 'Scripts') / 'twine'
        command = f"{twine_path} upload --verbose {upload_dir}/*"
        try:
            subprocess.run(command.split(), check=True)
            print(f"{_.GREEN}Done.{_.RESET}")
        except subprocess.CalledProcessError:
            print(f"{_.RED}Failed{_.RESET} to upload packages to PYPI.")
            sys.exit(1)
        
        sys.exit(0)


from colorama import Fore as f
from os.path import join
from os import listdir
from typing import List
import subprocess, re

from prompt_toolkit import PromptSession, HTML
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import WordCompleter, PathCompleter, Completer, merge_completers

class _InteractiveShell:
    """Interactive Shell Wrap"""
    def __init__(self, bin_path: str) -> None:
        """`create shell wrapper object`"""
        ...
    
    def _execute(self, command: List[str]) -> None:
        """`run command`"""
        ...
    
    def _strip_ansi(self, text: str) -> str:
        """`stripped ansi`"""
        ...
    
    def _get_completer(self) -> Completer:
        """merge and return completers"""
        ...
    
    def start_shell(self, prompt: HTML, stop: List[str]) -> None:
        """start the interactive shell.
        
        #### Parameters

        - `prompt`: This is <class 'prompt_toolkit.HTML'>
        - `stop`: List of stop keywords
        """
        ...

class TerminalEmulator:
    """`Terminal Emulator for Enviroment`"""
    def __init__(self, env: str, bin_path: str) -> None:
        """Create a Terminal Emulator Object"""
        ...
    
    @property
    def start(self) -> None:
        """`Start`"""
        ...


from .definitions import ArgumentDefinition
from argpi import Arguments as argpi, ArgumentDescription

class Arguments:
    def __init__(self) -> None:
        self.arguments = argpi().__capture__()

        # add
        for long, short in ArgumentDefinition.__arguments__.items():
            self.arguments.__add__(long, ArgumentDescription().shorthand(short))
        
        # analyse
        self.arguments.__analyse__()
    
    @property
    def analyse_and_get(self) -> argpi:
        return self.arguments


class TomlDirectoryError(Exception):
    NOT_A_DIR = "NOT A DIRECTORY: {} (internal error)"

class TomlFileError(Exception):
    FILE_NOT_FOUND = "\'cleverish.toml\' or \'Cleverish.toml\' not found. Refer to official documentation for help."
    NOT_A_FILE = "{} is not a FILE."

class TomlerError(Exception):
    pass


from .tomler_ import Tomler
from .configure_ import ConfigurationInterpreter
from .emulator import TerminalEmulator

__all__ = [
    "Tomler",
    "ConfigurationInterpreter",
    "TerminalEmulator",
]


class NonEmptyDirectory(Exception):
    NOT_EMPTY = "{} directory is not empty."

class CleverConfigurationNotFound(Exception):
    INIT_FIRST = "Clever config (clever.toml) is not present in current directory, please run clever init."


class TomlDirectoryError(Exception):
    """`Toml Config directory error.`"""
    NOT_A_DIR: str

class TomlFileError(Exception):
    """`Toml file not found Error and related errors`"""
    FILE_NOT_FOUND: str
    NOT_A_FILE: str

class TomlerError(Exception):
    """`Generic Tomler Error`"""



from typing import List, Dict
from json import dump as dump_as_json, load as load_from_json

class DotLock:
    """`.lock Class`"""
    @staticmethod
    def create(dependencies: List[Dict[str, str]], name: str = 'clever') -> None:
        """`Create the .lock with given dependencies`"""
        ...
    
    @staticmethod
    def load(name: str = 'clever') -> List[Dict[str, str]]:
        """`Load the .lock file with given name.`"""
        ...



from .tomler_ import Tomler
from ..exceptions.configuration_exceptions import ConfigurationError
from typing import Dict, List
import os

class _ConfigurationObjectives:
    def __init__(self, configuration_object: Tomler) -> None:
        self.object = configuration_object
        self.get
    
    @property
    def get(self) -> None:
        # Try to get environment_name
        self._environment_name = self.object.fetch_entry('config.environment_name')

        if self._environment_name is None:
            self._environment_name = os.path.basename(os.path.dirname(self.object.filepath))
            self.object.add_or_modify_entry('config.environment_name', self.environment_name)
        
        if not isinstance(self._environment_name, str):
            raise ConfigurationError(f"Expected <class 'str'> for [config.environment_name], found: {type(self._environment_name)}")
        
        # Try to get dependency_directory
        self._dependency_directory = self.object.fetch_entry('requirements.dependency_directory')

        if self._dependency_directory is None:
            self._dependency_directory = os.path.join(os.path.dirname(self.object.filepath), 'deps')
            self.object.add_or_modify_entry('requirements.dependency_directory', self.dependency_directory)
        
        if not isinstance(self._dependency_directory, str):
            raise ConfigurationError(f"Expected <class 'str'> for [requirements.dependency_directory], found: {type(self._dependency_directory)}")
        
        # Try to get get-automatically
        self._requirements_auto = self.object.fetch_entry('requirements.get-automatically')

        if self._requirements_auto is None:
            self._requirements_auto = False
            self.object.add_or_modify_entry('requirements.get-automatically', self.get_requirements_automatically)
        
        if not isinstance(self._requirements_auto, bool):
            raise ConfigurationError(f"Expected <class 'bool'> for [requirements.get-automatically], found: {type(self._requirements_auto)}")
        
        # Try to get list of requirements
        self._requirements_list = self.object.fetch_entry('requirements.list')

        if self._requirements_list is None:
            self._requirements_list: List[Dict[str, str]] = []
            self.object.add_or_modify_entry('requirements.list', self.requirements_list)
        
        if not isinstance(self._requirements_list, list):
            raise ConfigurationError(f"Expected <class 'builtins.list'> for [requirements.list], found: {type(self._requirements_list)}")
    
    @property
    def reload(self) -> None:
        self.object.load_configuration
        self.get
    
    @property
    def environment_name(self) -> str:
        return self._environment_name

    @property
    def dependency_directory(self) -> str:
        return self._dependency_directory

    @property
    def get_requirements_automatically(self) -> bool:
        return self._requirements_auto
    
    @property
    def requirements_list(self) -> List[Dict[str, str]]:
        return self._requirements_list

class ConfigurationInterpreter:
    @staticmethod
    def interpret(tomler: Tomler) -> _ConfigurationObjectives:
        return _ConfigurationObjectives(tomler)
    